@@ -1,18 +1,30 @@
-import subprocess
-import shlex
-from .command import SiteCommand, Fail
-from staticsite.page_filter import PageFilter
+from __future__ import annotations
+
+import argparse
 import logging
+import shlex
+import subprocess
+from typing import TYPE_CHECKING, Any, Optional, cast
+
+from staticsite.page_filter import PageFilter
+from staticsite.features.taxonomy import TaxonomyFeature
+
+from .command import Fail, SiteCommand, register
+
+if TYPE_CHECKING:
+    from ..page import SourcePage
+
 
 log = logging.getLogger("edit")
 
 
+@register
 class Edit(SiteCommand):
     "edit an existing page. Bare keywords match titles and file names, '+tag' matches tags"
 
     MENU_SIZE = 5
 
-    def select_page_menu(self, pages):
+    def select_page_menu(self, pages: list[SourcePage]) -> Optional[SourcePage]:
         """
         Prompt the user to select a page from `pages`.
 
@@ -60,15 +72,15 @@ class Edit(SiteCommand):
                 else:
                     return pages[idx - 1]
 
-    def run(self):
+    def run(self) -> Optional[int]:
         site = self.load_site()
 
-        filter_args = {
+        filter_args: dict[str, Any] = {
             "sort": "-date",
         }
 
         # Filter by taxonomy
-        taxonomies = site.features["taxonomy"]
+        taxonomies: TaxonomyFeature = cast(TaxonomyFeature, site.features["taxonomy"])
         args = []
         for arg in self.args.match:
             if arg.startswith("+"):
@@ -81,7 +93,7 @@ class Edit(SiteCommand):
                 else:
                     args.append(arg)
 
-        def match_page(page):
+        def match_page(page: SourcePage) -> bool:
             for m in args:
                 if (m.lower() not in page.meta.get("title", "").lower()
                         and m.lower() not in page.src.relpath.lower()):
@@ -90,18 +102,20 @@ class Edit(SiteCommand):
 
         # Build a list of all findable pages, present on disk, sorted with the newest first
         f = PageFilter(site, **filter_args)
-        pages = [page for page in f.filter(site.pages.values()) if page.src.stat is not None and match_page(page)]
+        pages = [page for page in f.filter()
+                 if isinstance(page, SourcePage) and match_page(page)]
 
         if len(pages) == 0:
             raise Fail("No page found matching {}".format(" ".join(shlex.quote(x) for x in self.args.match)))
 
+        page: Optional[SourcePage]
         if len(pages) == 1:
             page = pages[0]
         else:
             page = self.select_page_menu(pages)
 
         if page is None:
-            return
+            return None
 
         abspath = page.src.abspath
 
@@ -114,10 +128,11 @@ class Edit(SiteCommand):
                 log.warn("Editor command %s exited with error %d", " ".join(shlex.quote(x) for x in cmd), e.returncode)
                 return e.returncode
         print(abspath)
+        return None
 
     @classmethod
-    def make_subparser(cls, subparsers):
-        parser = super().make_subparser(subparsers)
+    def add_subparser(cls, subparsers: "argparse._SubParsersAction[Any]") -> argparse.ArgumentParser:
+        parser = super().add_subparser(subparsers)
         parser.add_argument("match", nargs="*", help="keywords used to look for the page to edit")
         parser.add_argument("-n", "--noedit", action="store_true",
                             help="do not run an editor, only output the file name of the new post")
