@@ -1,8 +1,16 @@
-import os
-import subprocess
-import shlex
-from .command import SiteCommand, Fail
+from __future__ import annotations
+
+import argparse
 import logging
+import os
+import shlex
+import subprocess
+from typing import TYPE_CHECKING, Any, Optional
+
+from .command import Fail, SiteCommand, register
+
+if TYPE_CHECKING:
+    from ..site import Site
 
 log = logging.getLogger("new")
 
@@ -14,15 +22,15 @@ class LazyTitle:
     If no value has been provided and one is requested by the template, ask the
     user for one.
     """
-    def __init__(self, title=None):
+    def __init__(self, title: Optional[str] = None) -> None:
         self.title = title
 
-    def __str__(self):
+    def __str__(self) -> str:
         if self.title is None:
             self.title = input("Please enter the post title: ")
         return self.title
 
-    def __len__(self):
+    def __len__(self) -> int:
         return len(str(self))
 
 
@@ -30,18 +38,19 @@ class LazySlug:
     """
     Generate a slug from the value of LazyTitle
     """
-    def __init__(self, site, lazy_title):
+    def __init__(self, site: Site, lazy_title: LazyTitle):
         self.site = site
         self.lazy_title = lazy_title
 
-    def __str__(self):
+    def __str__(self) -> str:
         return self.site.slugify(str(self.lazy_title))
 
 
+@register
 class New(SiteCommand):
     "create a new page"
 
-    def run(self):
+    def run(self) -> Optional[int]:
         site = self.load_site()
 
         archetype = site.archetypes.find(self.args.archetype)
@@ -57,6 +66,11 @@ class New(SiteCommand):
         if relpath is None:
             raise Fail("archetype {} does not contain `path` in its front matter".format(archetype.relpath))
 
+        if site.settings.PROJECT_ROOT is None:
+            raise Fail("PROJECT_ROOT is empty")
+        if site.settings.CONTENT is None:
+            raise Fail("CONTENT is empty")
+
         content_root = os.path.join(site.settings.PROJECT_ROOT, site.settings.CONTENT)
         abspath = os.path.join(content_root, relpath)
         if self.args.overwrite or not os.path.exists(abspath):
@@ -75,10 +89,11 @@ class New(SiteCommand):
                 log.warn("Editor command %s exited with error %d", " ".join(shlex.quote(x) for x in cmd), e.returncode)
                 return e.returncode
         print(abspath)
+        return None
 
     @classmethod
-    def make_subparser(cls, subparsers):
-        parser = super().make_subparser(subparsers)
+    def add_subparser(cls, subparsers: "argparse._SubParsersAction[Any]") -> argparse.ArgumentParser:
+        parser = super().add_subparser(subparsers)
         parser.add_argument("-a", "--archetype", default="default", help="page archetype")
         parser.add_argument("-t", "--title", help="page title")
         parser.add_argument("-n", "--noedit", action="store_true",
