@@ -1,17 +1,26 @@
 from __future__ import annotations
-from typing import List, Optional, Union, Sequence, Dict, Set
-import jinja2
-import os
-import re
+
 import datetime
 import logging
+import os
+import re
 from collections import defaultdict
-from .page import Page, PageNotFoundError
-from .utils import front_matter, arrange
-from .utils.typing import Meta
-from .metadata import Metadata
-from .file import File
+from typing import (TYPE_CHECKING, Any, Callable, List, MutableMapping,
+                    Optional, Sequence, Type, Union)
+
+import jinja2
+import jinja2.sandbox
+import markupsafe
+
 from . import toposort
+from .file import File
+from .page import ImagePage, Page, PageNotFoundError
+from .utils import front_matter
+from .utils.arrange import arrange
+
+if TYPE_CHECKING:
+    from .page import Pages
+    from .site import Site
 
 log = logging.getLogger("theme")
 
@@ -28,11 +37,11 @@ class Loader:
         # Sequence of search paths to use to resolve theme names
         self.search_paths = search_paths
         # Configurations by name
-        self.configs: Dict[str, Meta] = {}
+        self.configs: dict[str, dict[str, Any]] = {}
         # Dependency graph of themes
-        self.deps: Dict[str, Set[str]] = defaultdict(set)
+        self.deps: dict[str, set[str]] = defaultdict(set)
 
-    def load(self, name: str) -> List[Meta]:
+    def load(self, name: str) -> list[dict[str, Any]]:
         """
         Load the configuration of the given theme and all its dependencies.
 
@@ -43,7 +52,7 @@ class Loader:
         sorted_names = toposort.sort(self.deps)
         return [self.configs[name] for name in sorted_names]
 
-    def load_legacy(self, path: str) -> List[Meta]:
+    def load_legacy(self, path: str) -> List[dict[str, Any]]:
         """
         Same as load, but start with a path to an initial theme
         """
@@ -54,12 +63,12 @@ class Loader:
 
         for parent in config["extends"]:
             self.load_configs(parent)
-            self.deps[parent].add(name)
+            self.deps[name].add(parent)
 
         sorted_names = toposort.sort(self.deps)
         return [self.configs[name] for name in sorted_names]
 
-    def load_configs(self, name: str):
+    def load_configs(self, name: str) -> None:
         """
         Populate self.configs with the configuration of the named theme,
         preceded by all its dependencies
@@ -74,7 +83,7 @@ class Loader:
 
         for parent in config["extends"]:
             self.load_configs(parent)
-            self.deps[parent].add(name)
+            self.deps[name].add(parent)
 
     def find_root(self, name: str) -> str:
         """
@@ -87,13 +96,13 @@ class Loader:
 
         raise ThemeNotFoundError(f"Theme {name!r} not found in {self.search_paths!r}")
 
-    def load_config(self, root: str, name: str) -> Meta:
+    def load_config(self, root: str, name: str) -> dict[str, Any]:
         """
         Load the configuration for the given named theme
         """
         pathname = os.path.join(root, "config")
         if not os.path.isfile(pathname):
-            config = {}
+            config: dict[str, Any] = {}
         else:
             with open(pathname, "rt") as fd:
                 fmt, config = front_matter.read_whole(fd)
@@ -117,18 +126,21 @@ class Loader:
 class Jinja2TemplateLoader(jinja2.loaders.BaseLoader):
     re_content = re.compile(r"^content:(.+)")
 
-    def __init__(self, theme):
+    def __init__(self, theme: Theme):
         self.loader_content = jinja2.FileSystemLoader(theme.site.content_root)
         self.loader_theme = jinja2.FileSystemLoader(theme.template_lookup_paths)
 
-    def get_loader(self, template):
+    def get_loader(self, template: str) -> tuple[jinja2.loaders.BaseLoader, str]:
         mo = self.re_content.match(template)
         if mo:
             return self.loader_content, mo.group(1)
         else:
             return self.loader_theme, template
 
-    def get_source(self, environment, template):
+    def get_source(
+            self,
+            environment: jinja2.Environment,
+            template: str) -> tuple[str, Optional[str], Optional[Callable[[], bool]]]:
         loader, name = self.get_loader(template)
         try:
             return loader.get_source(environment, name)
@@ -138,7 +150,11 @@ class Jinja2TemplateLoader(jinja2.loaders.BaseLoader):
             raise jinja2.TemplateNotFound(template)
 
     @jinja2.utils.internalcode
-    def load(self, environment, name, globals=None):
+    def load(
+            self,
+            environment: jinja2.Environment,
+            name: str,
+            globals: Optional[MutableMapping[str, Any]] = None) -> jinja2.Template:
         loader, local_name = self.get_loader(name)
         try:
             return loader.load(environment, local_name, globals)
@@ -147,7 +163,7 @@ class Jinja2TemplateLoader(jinja2.loaders.BaseLoader):
             # (the one that includes the prefix)
             raise jinja2.TemplateNotFound(name)
 
-    def list_templates(self):
+    def list_templates(self) -> list[str]:
         result = []
         for template in self.loader_theme.list_templates():
             result.append(template)
@@ -157,7 +173,7 @@ class Jinja2TemplateLoader(jinja2.loaders.BaseLoader):
 
 
 class Theme:
-    def __init__(self, site, name, configs: List[Meta]):
+    def __init__(self, site: Site, name: str, configs: list[dict[str, Any]]):
         # Site object
         self.site = site
 
@@ -169,10 +185,7 @@ class Theme:
         self.configs = configs
 
         # Jinja2 Environment
-        self.jinja2 = None
-
-        # Cached list of metadata that are templates for other metadata
-        self.metadata_templates: Optional[List[Metadata]] = None
+        self.jinja2: jinja2.Environment
 
         # Compute template lookup paths
         self.template_lookup_paths: List[str] = []
@@ -204,16 +217,19 @@ class Theme:
 
         # Merge theme metadata
         meta_keys = frozenset(("image_sizes",))
-        self.meta: Meta = {}
+        self.meta: dict[str, Any] = {}
         for config in self.configs:
             for key in config.keys() & meta_keys:
                 self.meta[key] = config[key]
 
     @classmethod
-    def create(cls, site, name: str, search_paths: Sequence[str] = None) -> "Theme":
+    def create(cls, site: Site, name: str, search_paths: Optional[Sequence[str]] = None) -> "Theme":
         """
         Create a Theme looking up its name in the theme search paths
         """
+        if site.settings.PROJECT_ROOT is None:
+            raise RuntimeError("PROJECT_ROOT is None")
+
         if search_paths is None:
             search_paths = [
                 os.path.join(site.settings.PROJECT_ROOT, path) for path in site.settings.THEME_PATHS
@@ -222,10 +238,13 @@ class Theme:
         return cls(site, name, loader.load(name))
 
     @classmethod
-    def create_legacy(cls, site, paths: Sequence[str]) -> "Theme":
+    def create_legacy(cls, site: Site, paths: Sequence[str]) -> "Theme":
         """
         Create a theme from a list of possible theme paths
         """
+        if site.settings.PROJECT_ROOT is None:
+            raise RuntimeError("PROJECT_ROOT is None")
+
         loader = Loader(site.settings.THEME_PATHS)
 
         for root in paths:
@@ -235,19 +254,18 @@ class Theme:
 
         raise ThemeNotFoundError(f"Theme not found in {paths!r}")
 
-    def load(self):
+    def load(self) -> None:
         # Load feature plugins from the theme directories
         self.site.features.load_feature_dir(self.feature_dirs)
         self.site.features.commit()
         self.site.stage_features_constructed = True
 
+        env_cls: Type[jinja2.Environment]
         # Jinja2 template engine
         if self.site.settings.JINJA2_SANDBOXED:
-            from jinja2.sandbox import ImmutableSandboxedEnvironment
-            env_cls = ImmutableSandboxedEnvironment
+            env_cls = jinja2.sandbox.ImmutableSandboxedEnvironment
         else:
-            from jinja2 import Environment
-            env_cls = Environment
+            env_cls = jinja2.Environment
 
         self.jinja2 = env_cls(
             loader=Jinja2TemplateLoader(self),
@@ -277,66 +295,54 @@ class Theme:
         )
 
         self.jinja2.filters["datetime_format"] = self.jinja2_datetime_format
+        self.jinja2.filters["next_month"] = self.jinja2_next_month
         self.jinja2.filters["basename"] = self.jinja2_basename
-        self.jinja2.filters["arrange"] = arrange
+        self.jinja2.filters["arrange"] = self.jinja2_arrange
 
         # Add feature-provided globals and filters
         for feature in self.site.features.ordered():
             self.jinja2.globals.update(feature.j2_globals)
             self.jinja2.filters.update(feature.j2_filters)
 
-    def scan_assets(self):
+    def scan_assets(self) -> None:
         """
         Load static assets
         """
-        meta = dict(self.site.content_roots[0].meta)
-        meta["asset"] = True
-        meta["site_path"] = site_path = os.path.join(meta["site_path"], self.site.settings.STATIC_PATH)
-
         # Load system assets from site settings and theme configurations
         for name in self.system_assets:
             root = os.path.join("/usr/share/javascript", name)
             if not os.path.isdir(root):
                 log.warning("%s: system asset directory not found", root)
                 continue
-            fmeta = dict(meta)
-            fmeta["site_path"] = os.path.join(site_path, name)
-            # TODO: make this a child of the previously scanned static
-            self.site.scan_tree(
-                src=File.with_stat(name, root),
-                meta=fmeta,
+            src = File.with_stat(name, root)
+            self.site.scan_asset_tree(
+                src=src,
+                node=self.site.static_root.asset_child(name, src=src),
             )
 
         # Load assets from theme directories
         for path in self.theme_static_dirs:
-            self.site.scan_tree(
+            self.site.scan_asset_tree(
                 src=File.with_stat("", os.path.abspath(path)),
-                meta=meta,
+                node=self.site.static_root,
             )
 
-    def precompile_metadata_templates(self, meta: Meta):
-        """
-        Precompile all the elements of the given metadata that are jinja2
-        template strings
-        """
-        if self.metadata_templates is None:
-            self.metadata_templates = [m for m in self.site.metadata.values() if m.type == "jinja2"]
-        for metadata in self.metadata_templates:
-            val = meta.get(metadata.name)
-            if val is None:
-                continue
-            if isinstance(val, str):
-                meta[metadata.name] = self.jinja2.from_string(val)
-
     def jinja2_basename(self, val: str) -> str:
         return os.path.basename(val)
 
-    @jinja2.contextfilter
-    def jinja2_datetime_format(self, context, dt: datetime.datetime, format: str = None) -> str:
+    @jinja2.pass_context
+    def jinja2_datetime_format(
+            self,
+            context: jinja2.runtime.Context,
+            dt: Union[str, datetime.datetime],
+            format: Optional[str] = None) -> str:
         if not isinstance(dt, datetime.datetime):
             import dateutil.parser
             dt = dateutil.parser.parse(dt)
-        if format in ("rss2", "rfc822"):
+        if format is None or format == "iso8601":
+            from .utils import format_date_iso8601
+            return format_date_iso8601(dt)
+        elif format in ("rss2", "rfc822"):
             from .utils import format_date_rfc822
             return format_date_rfc822(dt)
         elif format in ("atom", "rfc3339"):
@@ -345,9 +351,6 @@ class Theme:
         elif format == "w3cdtf":
             from .utils import format_date_w3cdtf
             return format_date_w3cdtf(dt)
-        elif format == "iso8601" or not format:
-            from .utils import format_date_iso8601
-            return format_date_iso8601(dt)
         elif format[0] == '%':
             return dt.strftime(format)
         else:
@@ -355,8 +358,24 @@ class Theme:
                      context.parent["page"].src.relpath, context.name, format)
             return "(unknown datetime format {})".format(format)
 
-    @jinja2.contextfunction
-    def jinja2_has_page(self, context, arg: str) -> bool:
+    @jinja2.pass_context
+    def jinja2_next_month(self, context: jinja2.runtime.Context, dt: Any) -> Union[datetime.date, datetime.datetime]:
+        if isinstance(dt, str):
+            import dateutil.parser
+            dt = dateutil.parser.parse(dt)
+
+        if isinstance(dt, datetime.datetime):
+            return (dt.replace(day=1) + datetime.timedelta(days=40)).replace(
+                    day=1, hour=0, minute=0, second=0, microsecond=0)
+        elif isinstance(dt, datetime.date):
+            return (dt.replace(day=1) + datetime.timedelta(days=40)).replace(day=1)
+        else:
+            raise RuntimeError(
+                    f"{context.parent['page']}+{context.name}: invalid datetime {dt!r} of type {type(dt)}:"
+                    " accepted are str, datetime, date")
+
+    @jinja2.pass_context
+    def jinja2_has_page(self, context: jinja2.runtime.Context, arg: str) -> bool:
         cur_page = context.get("page")
         try:
             cur_page.resolve_path(arg)
@@ -365,68 +384,86 @@ class Theme:
         else:
             return True
 
-    @jinja2.contextfunction
-    def jinja2_page_for(self, context, arg: Union[str, Page]) -> Page:
+    @jinja2.pass_context
+    def jinja2_page_for(self, context: jinja2.runtime.Context, arg: Union[str, Page]) -> Page:
         """
         Generate a URL for a page, specified by path or with the page itself
         """
         if isinstance(arg, Page):
             return arg
 
-        cur_page = context.get("page")
+        cur_page: Optional[Page] = context.get("page")
         if cur_page is None:
-            log.warn("%s+%s: page_for(%s): current page is not defined", cur_page, context.name, arg)
-            return ""
+            raise RuntimeError(f"{cur_page}+{context.name}: page_for({arg!r}): current page is not defined")
 
         try:
             return cur_page.resolve_path(arg)
         except PageNotFoundError as e:
-            log.warn("%s:%s: %s", cur_page, context.name, e)
-            return ""
+            raise RuntimeError(f"{cur_page}+{context.name}: {e}")
 
-    @jinja2.contextfunction
-    def jinja2_url_for(self, context, arg: Union[str, Page], absolute=False) -> str:
+    @jinja2.pass_context
+    def jinja2_url_for(
+            self,
+            context: jinja2.runtime.Context,
+            arg: Union[str, Page],
+            absolute: bool = False,
+            static: bool = False) -> str:
         """
         Generate a URL for a page, specified by path or with the page itself
         """
-        cur_page = context.get("page")
+        cur_page: Optional[Page] = context.get("page")
+        # print(f"Theme.jinja2_url_for {cur_page=!r}")
         if cur_page is None:
             log.warn("%s+%s: url_for(%s): current page is not defined", cur_page, context.name, arg)
             return ""
 
         try:
-            return cur_page.url_for(arg, absolute=absolute)
+            return cur_page.url_for(arg, absolute=absolute, static=static)
         except PageNotFoundError as e:
             log.warn("%s:%s: %s", cur_page, context.name, e)
             return ""
 
-    @jinja2.contextfunction
-    def jinja2_site_pages(self, context, **kw) -> List[Page]:
-        cur_page = context.get("page")
+    @jinja2.pass_context
+    def jinja2_site_pages(self, context: jinja2.runtime.Context, **kw: Any) -> list[Page]:
+        cur_page: Optional[Page] = context.get("page")
         if cur_page is None:
             log.warn("%s+%s: site_pages: current page is not defined", cur_page, context.name)
             return []
 
         return cur_page.find_pages(**kw)
 
-    @jinja2.contextfunction
+    @jinja2.pass_context
     def jinja2_img_for(
-            self, context,
+            self, context: jinja2.runtime.Context,
             path: Union[str, Page],
             type: Optional[str] = None,
             absolute: bool = False,
-            **attrs) -> List[Page]:
+            **attrs: Any) -> str:
         cur_page = context.get("page")
         if cur_page is None:
             log.warn("%s+%s: img(%s): current page is not defined", cur_page, context.name, path)
             return ""
 
-        res_attrs = cur_page.get_img_attributes(path, type=type, absolute=absolute)
-        res_attrs.update(attrs)
+        image_page = cur_page.resolve_path(path)
+        if isinstance(image_page, ImagePage):
+            res_attrs = image_page.get_img_attributes(type=type, absolute=absolute)
+        else:
+            log.warning("%s: img src= resolves to %s which is not an image page", cur_page, image_page)
+            res_attrs = {"src": cur_page.url_for(image_page)}
 
-        escape = jinja2.escape
+        escape = markupsafe.escape
         res = ["<img"]
         for k, v in res_attrs.items():
             res.append(f" {escape(k)}='{escape(v)}'")
         res.append("></img>")
-        return jinja2.Markup("".join(res))
+        return markupsafe.Markup("".join(res))
+
+    def jinja2_arrange(
+            self,
+            pages: Union[Pages, list[Page]],
+            *args: Any, **kw: Any) -> list[Page]:
+        from .page import Pages
+        if isinstance(pages, Pages):
+            return pages.arrange(*args, **kw)
+        else:
+            return arrange(pages, *args, **kw)
